@@ -20,7 +20,7 @@ import warnings
 from pathlib import Path
 from threading import local
 
-from IPython.core import page, payloadpage
+from IPython.core import page
 from IPython.core.autocall import ZMQExitAutocall
 from IPython.core.displaypub import DisplayPublisher
 from IPython.core.error import UsageError
@@ -33,12 +33,17 @@ from IPython.utils import openpy
 from IPython.utils.process import arg_split, system  # type:ignore[attr-defined]
 from jupyter_client.session import Session, extract_header
 from jupyter_core.paths import jupyter_runtime_dir
-from traitlets import Any, CBool, CBytes, Dict, Instance, Type, default, observe
+from traitlets import Any, Bool, CBool, CBytes, Dict, Instance, Type, default, observe
 
 from ipykernel import connect_qtconsole, get_connection_file, get_connection_info
 from ipykernel.displayhook import ZMQShellDisplayHook
 from ipykernel.jsonutil import encode_images, json_clean
 
+try:
+    from IPython.core.history import HistoryOutput
+except ImportError:
+    HistoryOutput = None  # type: ignore[assignment,misc]
+
 # -----------------------------------------------------------------------------
 # Functions and classes
 # -----------------------------------------------------------------------------
@@ -52,6 +57,11 @@ class ZMQDisplayPublisher(DisplayPublisher):
     parent_header = Dict({})
     topic = CBytes(b"display_data")
 
+    store_display_history = Bool(
+        False,
+        help="If set to True, store display outputs in the history manager. Default is False.",
+    ).tag(config=True)
+
     # thread_local:
     # An attribute used to ensure the correct output message
     # is processed. See ipykernel Issue 113 for a discussion.
@@ -100,6 +110,21 @@ class ZMQDisplayPublisher(DisplayPublisher):
         update : bool, optional, keyword-only
             If True, send an update_display_data message instead of display_data.
         """
+        if (
+            self.store_display_history
+            and self.shell is not None
+            and hasattr(self.shell, "history_manager")
+            and HistoryOutput is not None
+        ):
+            # Reference: github.com/ipython/ipython/pull/14998
+            exec_count = self.shell.execution_count
+            if getattr(self.shell.display_pub, "_in_post_execute", False):
+                exec_count -= 1
+            outputs = getattr(self.shell.history_manager, "outputs", None)
+            if outputs is not None:
+                outputs.setdefault(exec_count, []).append(
+                    HistoryOutput(output_type="display_data", bundle=data)
+                )
         self._flush_streams()
         if metadata is None:
             metadata = {}
@@ -302,7 +327,7 @@ class KernelMagics(Magics):
 
         # Make sure we send to the client an absolute path, in case the working
         # directory of client and kernel don't match
-        filename = Path(filename).resolve()
+        filename = str(Path(filename).resolve())
 
         payload = {"source": "edit_magic", "filename": filename, "line_number": lineno}
         assert self.shell is not None
@@ -508,10 +533,38 @@ class ZMQInteractiveShell(InteractiveShell):
         env["PAGER"] = "cat"
         env["GIT_PAGER"] = "cat"
 
+    def payloadpage_page(self, strg, start=0, screen_lines=0, pager_cmd=None):
+        """Print a string, piping through a pager.
+
+        This version ignores the screen_lines and pager_cmd arguments and uses
+        IPython's payload system instead.
+
+        Parameters
+        ----------
+        strg : str or mime-dict
+            Text to page, or a mime-type keyed dict of already formatted data.
+        start : int
+            Starting line at which to place the display.
+        """
+
+        # Some routines may auto-compute start offsets incorrectly and pass a
+        # negative value.  Offset to 0 for robustness.
+        start = max(0, start)
+
+        data = strg if isinstance(strg, dict) else {"text/plain": strg}
+
+        payload = dict(
+            source="page",
+            data=data,
+            start=start,
+        )
+        assert self.payload_manager is not None
+        self.payload_manager.write_payload(payload)
+
     def init_hooks(self):
         """Initialize hooks."""
         super().init_hooks()
-        self.set_hook("show_in_pager", page.as_hook(payloadpage.page), 99)
+        self.set_hook("show_in_pager", page.as_hook(self.payloadpage_page), 99)
 
     def init_data_pub(self):
         """Delay datapub init until request, for deprecation warnings"""
@@ -612,7 +665,7 @@ class ZMQInteractiveShell(InteractiveShell):
         """Initialize magics."""
         super().init_magics()
         self.register_magics(KernelMagics)
-        self.magics_manager.register_alias("ed", "edit")  # type:ignore[union-attr]
+        self.magics_manager.register_alias("ed", "edit")
 
     def init_virtualenv(self):
         """Initialize virtual environment."""
